"""
SMS_BP is a Python package for simulating single molecule localization microscopy experiments.
See the README.md file for more information.
GitHub: https://github.com/joemans3/SMS_BP
Pypi: https://pypi.org/project/SMS_BP/
    - pip install SMS_BP
Author: Baljyot Singh Parmar
Last updated: 2023-08-08

Features:
- fPALM / PALM
- sptPALM
- ...

"""

__version__ = "1.0.0"
